(* json-parser.sml
 *
 * COPYRIGHT (c) 2008 The Fellowship of SML/NJ (http://www.smlnj.org)
 * All rights reserved.
 *)

structure JSONParser : sig

    val parse : TextIO.instream -> JSON.value

    val parseFile : string -> JSON.value

  end = struct

    structure Lex = JSONLexer
    structure T = JSONTokens
    structure J = JSON

    fun parse' (srcMap, inStrm) = let
	  fun error (pos, msg, tok) = raise Fail(concat[
		  "error ", AntlrStreamPos.spanToString srcMap pos, ": ",
		  msg, ", found '", JSONTokens.toString tok, "'"
		])
	  val lexer = Lex.lex srcMap
	  fun parse_sequence (is_tok_end, parse_item) (strm : Lex.strm, itms) = let
		fun is_tok_sep tok = case tok of T.COMMA => true | _ => false
		val (tok, pos, strm') = lexer strm
		in if is_tok_sep tok
		   then error (pos, "parsing sequence", tok)
		   else if is_tok_end tok
			then (strm', itms)
			else let val (strm'', itm) = parse_item strm
				 val (tok', pos', strm''') = lexer strm''
			     in if is_tok_end tok'
				then (strm''', itm :: itms)
				else if is_tok_sep tok'
				     then let val (tok'', pos'', _) = lexer strm'''
					  in if is_tok_end tok''
					     then error (pos'', "parsing sequence", tok'')
					     else parse_sequence (is_tok_end, parse_item) (strm''', itm :: itms)
					  end
				     else error (pos', "parsing sequence", tok')
				 end
		end
	  fun parseValue (strm : Lex.strm) = let
		val (tok, pos, strm) = lexer strm
		in
		  case tok
		   of T.LB => parseArray strm
		    | T.LCB => parseObject strm
		    | T.KW_null => (strm, J.NULL)
		    | T.KW_true => (strm, J.BOOL true)
		    | T.KW_false => (strm, J.BOOL false)
		    | T.INT n => (strm, J.INT n)
		    | T.FLOAT f => (strm, J.FLOAT f)
		    | T.STRING s => (strm, J.STRING s)
		    | _ => error (pos, "parsing value", tok)
		  (* end case *)
		end
	  and parseArray (strm : Lex.strm) = let
		fun is_RB tok = case tok of T.RB => true | _ => false
		val (strm', elmnts) = parse_sequence (is_RB, parseValue) (strm, [])
		in (strm', J.ARRAY(List.rev elmnts))
		end
	  and parseObject (strm : Lex.strm) = let
		fun is_RCB tok = case tok of T.RCB => true | _ => false
		fun parse_field strm = let
			val (tok, pos, strm') = lexer strm
			in case tok
			    of T.STRING s =>
				(case lexer strm'
				  of (T.COLON, _, strm'') => let
					val (strm''', v) = parseValue strm''
					in (strm''', (s, v))
					end
				   | (tok', pos', _) => error (pos', "parsing field", tok')
				 (* end case *))
			     | _ => error (pos, "parsing field", tok)
			   (* end case *)
			end
		val (strm', flds) = parse_sequence (is_RCB, parse_field) (strm, [])
		in (strm', J.OBJECT(List.rev flds))
		end
	  in
	    #2 (parseValue (Lex.streamifyInstream inStrm))
	  end

    fun parse inStrm = parse' (AntlrStreamPos.mkSourcemap (), inStrm)

    fun parseFile fileName = let
	  val inStrm = TextIO.openIn fileName
	  val v = parse' (AntlrStreamPos.mkSourcemap' fileName, inStrm)
		handle ex => (TextIO.closeIn inStrm; raise ex)
	  in
	    TextIO.closeIn inStrm;
	    v
	  end

  end

